from typing import NamedTuple
import io
from PIL import Image
import logging
logging.getLogger('PIL').setLevel(logging.WARNING)


class Image32(NamedTuple):
    data: bytes
    width: int
    height: int

    @staticmethod
    def load(src: bytes) -> 'Image32':
        image = Image.open(io.BytesIO(src))
        if image.mode != 'RGBA':
            image = image.convert('RGBA')
        data = image.tobytes('raw')
        return Image32(data, image.width, image.height)


class Texture:
    def __init__(self, name: str, image: Image32) -> None:
        self.name = name
        self.image = image


class CubeMap(NamedTuple):
    xp: Image32
    xn: Image32
    yp: Image32
    yn: Image32
    zp: Image32
    zn: Image32



from OpenGL import GL
from glglue.ctypesmath.camera import Camera
import imgui.integrations.opengl
import imgui
from glglue.basecontroller import BaseController
from logging import getLogger
from .samplecontroller import Scene

logger = getLogger(__name__)


class ImGuiController(BaseController):
    """
    [CLASSES] Controllerクラスは、glglueの規約に沿って以下のコールバックを実装する
    """

    def __init__(self, scale=1):
        self.is_init = False
        imgui.create_context()
        self.io = imgui.get_io()
        self.io.fonts.get_tex_data_as_rgba32()
        self.io.display_fb_scale = scale, scale
        self.viewport = (1, 1)
        self.camera = Camera()
        self.scene = Scene()

    def onResize(self, w, h):
        if self.viewport == (w, h):
            return False
        self.camera.onResize(w, h)
        self.viewport = (w, h)
        return True

    def onLeftDown(self, x, y):
        self.io.mouse_down[0] = 1
        if not self.io.want_capture_mouse:
            self.camera.onLeftDown(x, y)

        return False

    def onLeftUp(self, x, y):
        self.io.mouse_down[0] = 0
        if not self.io.want_capture_mouse:
            self.camera.onLeftUp(x, y)

        return False

    def onMiddleDown(self, x, y):
        self.io.mouse_down[2] = 1
        if not self.io.want_capture_mouse:
            self.camera.onMiddleDown(x, y)

        return False

    def onMiddleUp(self, x, y):
        self.io.mouse_down[2] = 0
        if not self.io.want_capture_mouse:
            self.camera.onMiddleUp(x, y)

        return False

    def onRightDown(self, x, y):
        self.io.mouse_down[1] = 1
        if not self.io.want_capture_mouse:
            self.camera.onRightDown(x, y)

        return False

    def onRightUp(self, x, y):
        self.io.mouse_down[1] = 0
        if not self.io.want_capture_mouse:
            self.camera.onRightUp(x, y)

        return False

    def onMotion(self, x, y):
        self.io.mouse_pos = x, y
        if not self.io.want_capture_mouse:
            self.camera.onMotion(x, y)

        return False

    def onWheel(self, d):
        self.io.mouse_wheel = d
        if not self.io.want_capture_mouse:
            self.camera.onWheel(d)

        return False

    def onKeyDown(self, keycode):
        logger.debug('onKeyDown: %d', keycode)

    def onUpdate(self, d):
        #logger.debug('onUpdate: delta %d ms', d)
        self.io.delta_time = d * 0.001
        return True

    def on_imgui(self):
        # open new window context
        imgui.begin("Your first window!", True)
        # draw text label inside of current window
        imgui.text("Hello world!")
        # close current window context
        imgui.end()

    def draw(self):
        if not self.is_init:
            # initilize imgui context (see documentation)
            self.imgui_impl = imgui.integrations.opengl.ProgrammablePipelineRenderer(
            )
            self.is_init = True

        state = self.camera.get_state()

        #
        # imgui
        #
        self.io.display_size = self.viewport

        # start new frame context
        imgui.new_frame()

        self.on_imgui()

        imgui.end_frame()

        #
        # render
        #
        GL.glViewport(0, 0, *self.viewport)

        GL.glClearColor(0.0, 0.0, 1.0, 0.0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT |
                   GL.GL_DEPTH_BUFFER_BIT)  # type: ignore
        if self.scene:
            self.scene.draw(state)

        # pass all drawing comands to the rendering pipeline
        # and close frame context
        imgui.render()
        self.imgui_impl.render(imgui.get_draw_data())

        GL.glFlush()



# -*- coding: utf-8 -*-
import logging
from PySide6 import QtCore, QtGui, QtOpenGLWidgets, QtWidgets
from .basecontroller import BaseController


class Widget(QtOpenGLWidgets.QOpenGLWidget):
    '''
    https://doc.qt.io/qtforpython/PySide6/QtOpenGLWidgets/QOpenGLWidget.html
    '''

    def __init__(self, parent, controller: BaseController, dpi_scale=1.0, core_profile=True):
        super().__init__(parent)
        self.controller = controller
        self.setMouseTracking(True)
        self.dpi_Scale = dpi_scale

        if core_profile:
            format = QtGui.QSurfaceFormat()
            format.setDepthBufferSize(24)
            format.setStencilBufferSize(8)
            format.setVersion(3, 2)
            format.setProfile(QtGui.QSurfaceFormat.CoreProfile)
            # self.setFormat(format)
            QtGui.QSurfaceFormat.setDefaultFormat(format)
            # must be called before the widget or its parent window gets shown

    def minimumSizeHint(self):
        return QtCore.QSize(50, 50)

    def sizeHint(self):
        return QtCore.QSize(150, 150)

    def paintGL(self):
        self.controller.draw()

    def resizeGL(self, width, height):
        if self.controller.onResize(int(width * self.dpi_Scale), int(height * self.dpi_Scale)):
            self.repaint()

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            if self.controller.onLeftDown(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
                self.repaint()
        elif event.button() == QtCore.Qt.MiddleButton:
            if self.controller.onMiddleDown(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
                self.repaint()
        elif event.button() == QtCore.Qt.RightButton:
            if self.controller.onRightDown(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
                self.repaint()

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            if self.controller.onLeftUp(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
                self.repaint()
        elif event.button() == QtCore.Qt.MiddleButton:
            if self.controller.onMiddleUp(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
                self.repaint()
        elif event.button() == QtCore.Qt.RightButton:
            if self.controller.onRightUp(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
                self.repaint()

    def mouseMoveEvent(self, event):
        if self.controller.onMotion(event.x() * self.dpi_Scale, event.y() * self.dpi_Scale):
            self.repaint()

    def wheelEvent(self, event):
        if self.controller.onWheel(-event.angleDelta().y()):
            self.repaint()


class CustomLogger(logging.Handler):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def emit(self, record):
        msg = self.format(record)

        if record.levelno == logging.DEBUG:
            msg = f'<font color="gray">{msg}</font><br>'
        elif record.levelno == logging.WARNING:
            msg = f'<font color="orange">{msg}</font><br>'
        elif record.levelno == logging.ERROR:
            msg = f'<font color="red">{msg}</font><br>'
        else:
            msg = f'{msg}<br>'

        self.widget.textCursor().movePosition(QtGui.QTextCursor.Start)
        self.widget.textCursor().insertHtml(msg)

    def write(self, m):
        pass

class QPlainTextEditLogger(QtWidgets.QPlainTextEdit):
    def __init__(self, parent):
        super().__init__(parent)
        self.setReadOnly(True)
        self.log_handler = CustomLogger(self)



import OpenGL.arrays.buffers
import ctypes
from typing import List, NamedTuple, Optional, Union
from OpenGL import GL
from ..scene.vertices import Planar, Interleaved, VectorView


class MemoryViewHandler:
    def __init__(self) -> None:
        pass

    @classmethod
    def asArray(cls, value: memoryview, type_code=None):
        match value.obj:
            case ctypes.Array() | bytes() | bytearray() as a:
                return a
            case _:
                # array.array
                if value.contiguous:
                    buffer = (ctypes.c_byte * value.nbytes).from_buffer(value)
                    return buffer
                else:
                    raise RuntimeError()

    @classmethod
    def from_param(cls, value: memoryview, typeCode=None):
        return cls.asArray(value)


# override builtins.memoryview handler
OpenGL.arrays.buffers.BufferHandler = MemoryViewHandler


class VertexAttribute(NamedTuple):
    offset: int
    stride: int
    # Vec2, Vec3, Vec4 などの2, 3, 4
    component_count: int


class VBO:
    def __init__(self) -> None:
        self.vbo = GL.glGenBuffers(1)
        self.vertex_count = 0
        self.attributes: List[VertexAttribute] = []

    def __del__(self) -> None:
        GL.glDeleteBuffers(1, [self.vbo])

    def bind(self) -> None:
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)

    def unbind(self) -> None:
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)

    def set_vertex_attribute(self, data: memoryview, stride: int, offsets: List[int], is_dynamic: bool) -> None:
        ''' float2, 3, 4'''
        self.stride = stride
        self.vertex_count = len(data) // stride
        self.bind()
        GL.glBufferData(GL.GL_ARRAY_BUFFER, data.nbytes, data,
                        GL.GL_DYNAMIC_DRAW if is_dynamic else GL.GL_STATIC_DRAW)
        self.unbind()
        if offsets:
            for i, offset in enumerate(offsets):
                end = stride
                if i < len(offsets)-1:
                    end = offsets[i+1]
                self.attributes.append(
                    VertexAttribute(offset, stride, (end-offset) // 4))
        else:
            self.attributes.append(VertexAttribute(0, stride, stride // 4))

    def update(self, data: memoryview) -> None:
        self.bind()
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER, 0, data.nbytes, data)
        self.unbind()

    def set_slot(self, slot: int) -> None:
        self.bind()
        for a in self.attributes:
            GL.glEnableVertexAttribArray(slot)
            GL.glVertexAttribPointer(slot, a.component_count, GL.GL_FLOAT, GL.GL_FALSE,
                                     a.stride,  ctypes.c_void_p(a.offset))
            slot += 1

    def draw(self) -> None:
        GL.glDrawArrays(GL.GL_TRIANGLES, 0, self.vertex_count)

    def draw_lines(self) -> None:
        GL.glDrawArrays(GL.GL_LINES, 0, self.vertex_count)


def create_vbo_from(src: Union[Planar, Interleaved], is_dynamic=False) -> List[VBO]:
    vbo_list = []
    match src:
        case Planar(attributes):
            for a in attributes:
                vbo = VBO()
                vbo.set_vertex_attribute(
                    a.data, a.get_stride(), [], is_dynamic)
                vbo_list.append(vbo)
        case Interleaved(vertices, offsets):
            vbo = VBO()
            vbo.set_vertex_attribute(
                vertices.data, vertices.get_stride(), offsets, is_dynamic)
            vbo_list.append(vbo)
        case _:
            raise RuntimeError()
    return vbo_list


class IBO:
    def __init__(self) -> None:
        self.vbo = GL.glGenBuffers(1)
        self.index_count = 0
        self.index_type = 0
        self.topology = GL.GL_TRIANGLES

    def __del__(self) -> None:
        GL.glDeleteBuffers(1, [self.vbo])

    def bind(self) -> None:
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.vbo)

    def unbind(self) -> None:
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0)

    def set_indices(self, data: memoryview, stride: int) -> None:
        self.index_count = len(data) // stride
        self.bind()
        if stride == 1:
            raise Exception("not implemented")
        elif stride == 2:
            self.index_type = GL.GL_UNSIGNED_SHORT
        elif stride == 4:
            self.index_type = GL.GL_UNSIGNED_INT
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER,
                        data.nbytes, data, GL.GL_STATIC_DRAW)
        self.unbind()

    def draw(self) -> None:
        self.bind()
        GL.glDrawElements(self.topology, self.index_count,
                          self.index_type, None)


def create_ibo_from(src: VectorView) -> IBO:
    ibo = IBO()
    ibo.set_indices(src.data, src.get_stride())
    return ibo


class VAO:
    '''
    https://wlog.flatlib.jp/item/1629
    '''

    def __init__(self, index_format=None):
        self.vao = GL.glGenVertexArrays(1)
        self.index_format = index_format

    def __del__(self) -> None:
        GL.glDeleteVertexArrays(1, [self.vao])

    def bind(self):
        GL.glBindVertexArray(self.vao)

    def unbind(self):
        GL.glBindVertexArray(0)

    def draw(self, topology, offset: int, draw_count: int):
        self.bind()
        if self.index_format:
            GL.glDrawElements(topology, draw_count,
                              self.index_format, ctypes.c_void_p(offset))
        else:
            GL.glDrawArrays(topology, offset, draw_count)
        self.unbind()


def create_vao_from(ibo: Optional[IBO], vbo_list: List[VBO]) -> VAO:
    if ibo:
        vao = VAO(ibo.index_type)
    else:
        vao = VAO()
    vao.bind()
    if ibo:
        ibo.bind()

    if len(vbo_list) == 1 and len(vbo_list[0].attributes) > 1:
        # interleaved
        vbo = vbo_list[0]
        vbo.bind()
        for i, a in enumerate(vbo.attributes):
            GL.glEnableVertexAttribArray(i)
            GL.glVertexAttribPointer(i, a.component_count, GL.GL_FLOAT, GL.GL_FALSE,
                                     a.stride,  ctypes.c_void_p(a.offset))
    else:
        # planar
        for i, vbo in enumerate(vbo_list):
            vbo.bind()
            vbo.set_slot(i)

    vao.unbind()
    return vao


class Drawable(NamedTuple):
    vbo_list: List[VBO]
    ibo: Optional[IBO]
    vao: VAO

    def draw(self, topology, offset: int, draw_count: int):
        self.vao.draw(topology, offset, draw_count)


def create(vertices: Union[Planar, Interleaved], indices: Optional[VectorView] = None, *, is_dynamic=False):
    vbo_list = create_vbo_from(vertices, is_dynamic=is_dynamic)
    ibo = None
    if indices:
        ibo = create_ibo_from(indices)
    vao = create_vao_from(ibo, vbo_list)
    return Drawable(vbo_list, ibo, vao)



import ctypes


class Quaternion(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_float),
        ('y', ctypes.c_float),
        ('z', ctypes.c_float),
        ('w', ctypes.c_float),
    ]
    __match_args__ = ('x', 'y', 'z', 'w')

    def __eq__(self, rhs) -> bool:
        match rhs:
            case Quaternion(x, y, z) | (x, y, z):
                return self.x == x and self.y == y and self.z == z
        return False

    def __add__(self, rhs) -> 'Quaternion':
        match rhs:
            case Quaternion(x, y, z) | (x, y, z):
                return Quaternion(self.x + x, self.y + y, self.z + z)
        raise RuntimeError()



from OpenGL import GL
from OpenGL.raw.GL.VERSION.GL_1_0 import glTexImage2D


import OpenGL.images
OpenGL.images.TYPE_TO_ARRAYTYPE[GL.GL_HALF_FLOAT] = GL.GL_UNSIGNED_SHORT


class Texture:
    def __init__(self) -> None:
        self.texture = GL.glGenTextures(1)

    def __del__(self) -> None:
        GL.glDeleteTextures(1, [self.texture])

    def bind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)

    def unbind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

    def load(self, data: bytes, width: int, height: int):
        self.bind()
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, width,
                        height, 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, data)
        GL.glTexParameterf(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        GL.glTexParameterf(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        self.unbind()

    def activate(self, location: int, slot: int):
        self.bind()
        GL.glActiveTexture(GL.GL_TEXTURE0 + slot)  # type: ignore
        GL.glUniform1i(location, slot)


class CubeMap:
    def __init__(self) -> None:
        self.texture = GL.glGenTextures(1)

    def bind(self):
        GL.glBindTexture(GL.GL_TEXTURE_CUBE_MAP, self.texture)

    def unbind(self):
        GL.glBindTexture(GL.GL_TEXTURE_CUBE_MAP, 0)

    def _load_cube_map_side(self, side_target, data: bytes, width: int, height: int):
        # copy image data into 'target' side of cube map
        GL.glTexImage2D(
            side_target,
            0,
            GL.GL_RGBA16F,
            width,
            height,
            0,
            GL.GL_RGBA,
            GL.GL_HALF_FLOAT,
            data)

    def load(self, xp, xn, yp, yn, zp, zn):
        self.bind()

        # load each image and copy into a side of the cube-map texture
        self._load_cube_map_side(
            GL.GL_TEXTURE_CUBE_MAP_POSITIVE_X, xp.data, xp.width, xp.height)
        self._load_cube_map_side(
            GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_X, xp.data, xp.width, xp.height)
        self._load_cube_map_side(
            GL.GL_TEXTURE_CUBE_MAP_POSITIVE_Y, yp.data, yp.width, yp.height)
        self._load_cube_map_side(
            GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, yp.data, yp.width, yp.height)
        self._load_cube_map_side(
            GL.GL_TEXTURE_CUBE_MAP_POSITIVE_Z, zp.data, zp.width, zp.height)
        self._load_cube_map_side(
            GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, zp.data, zp.width, zp.height)

        # format cube map texture
        GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP,
                           GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP,
                           GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP,
                           GL.GL_TEXTURE_WRAP_R, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP,
                           GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP,
                           GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE)

        self.unbind()



#!/usr/bin/python
# coding: utf-8
""" [NAME] glglue.glut

[DESCRIPTION] use mainloop(Controller)
"""

from OpenGL.GLUT import *
import time
from logging import getLogger
import sys
import pathlib

HERE = pathlib.Path(__file__).absolute().parent
sys.path.insert(0, str(HERE.parent))
logger = getLogger(__name__)

FPS = 30
MSPF = int(1000.0 / FPS)

_g_controller = None


def _resize(w, h):
    _g_controller.onResize(w, h)


def _mouse(button, state, x, y):
    if button == GLUT_LEFT_BUTTON:
        if state == GLUT_DOWN:
            if _g_controller.onLeftDown(x, y):
                glutPostRedisplay()
        else:
            if _g_controller.onLeftUp(x, y):
                glutPostRedisplay()
    elif button == GLUT_MIDDLE_BUTTON:
        if state == GLUT_DOWN:
            if _g_controller.onMiddleDown(x, y):
                glutPostRedisplay()
        else:
            if _g_controller.onMiddleUp(x, y):
                glutPostRedisplay()
    elif button == GLUT_RIGHT_BUTTON:
        if state == GLUT_DOWN:
            if _g_controller.onRightDown(x, y):
                glutPostRedisplay()
        else:
            if _g_controller.onRightUp(x, y):
                glutPostRedisplay()

    elif button == 3:
        if _g_controller.onWheel(-1):
            glutPostRedisplay()

    elif button == 4:
        if _g_controller.onWheel(1):
            glutPostRedisplay()

    else:
        logger.warn("unknown mouse:", button, state, x, y)


def _motion(x, y):
    if _g_controller.onMotion(x, y):
        glutPostRedisplay()


def _keyboard(key, x, y):
    if _g_controller.onKeyDown(ord(key)):
        glutPostRedisplay()


def _draw():
    _g_controller.draw()
    glutSwapBuffers()


def _timer(_):
    _g_controller.onUpdate(MSPF)
    glutTimerFunc(MSPF, _timer, 0)
    glutPostRedisplay()


def _create_idle_func():
    # nonlocal
    lastclock = [0]

    def idle():
        clock = time.clock()
        d = (clock - lastclock[0]) * 1000
        _g_controller.onUpdate(d)
        glutPostRedisplay()
        lastclock[0] = clock

    return idle


def setup(controller,
          width: int = 640,
          height: int = 480,
          title: bytearray = b"glut sample"):
    """ [FUNCTIONS] setup and start glut mainloop
    """
    global _g_controller
    _g_controller = controller

    glutInit(sys.argv)
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL)
    glutInitWindowSize(width, height)
    glutCreateWindow(title)
    # Windowのサイズが変わった時に呼ばれる関数を登録
    glutReshapeFunc(_resize)
    # 描画時に呼ばれる関数を登録
    glutDisplayFunc(_draw)
    # マウスボタン押し上げ時に呼ばれる関数
    glutMouseFunc(_mouse)
    # マウスドラッグ時に呼ばれる関数
    glutMotionFunc(_motion)
    # キーボードが押された時に呼ばれる関数
    glutKeyboardFunc(_keyboard)


def mainloop(controller,
             width: int = 640,
             height: int = 480,
             title: bytearray = b"glut sample"):
    """ [FUNCTIONS] setup and start glut mainloop
    """
    setup(controller, width, height, title)

    if glutMainLoopEvent:
        # manual loop
        lastclock = 0
        while True:
            glutMainLoopEvent()
            clock = time.perf_counter() * 1000
            d = clock - lastclock
            lastclock = clock
            _g_controller.onUpdate(d)
            _g_controller.draw()
            glutSwapBuffers()
    else:
        # タイマー
        #glutTimerFunc(MSPF, timer , 0);
        # idle
        glutIdleFunc(_create_idle_func())
        # start mainloop
        glutMainLoop()


class LoopManager:
    def __init__(self, controller, **kw):
        self.controller = controller
        setup(controller, **kw)

    def begin_frame(self):
        glutMainLoopEvent()
        clock = time.perf_counter() * 1000
        return clock

    def end_frame(self):
        glutSwapBuffers()


if __name__ == "__main__":
    import glglue.gl3

    if glutMainLoopEvent:
        # manual loop
        lastclock = 0
        loop = LoopManager(glglue.gl3.SampleController(),
                           width=480,
                           height=480,
                           title=b"glut")
        while True:
            clock = loop.begin_frame()
            d = clock - lastclock
            lastclock = clock
            _g_controller.onUpdate(d)
            _g_controller.draw()
            loop.end_frame()
    else:
        setup(glglue.sample.SampleController(),
              width=480,
              height=480,
              title=b"glut")
        # タイマー
        #glutTimerFunc(MSPF, timer , 0);
        # idle
        glutIdleFunc(_create_idle_func())
        # start mainloop
        glutMainLoop()



# coding: utf-8
if __name__ == "__main__":
    import glglue.gl3.samplecontroller
    controller = glglue.gl3.samplecontroller.SampleController()

    import glglue.glut
    glglue.glut.mainloop(controller)



from typing import Optional
from .texture import Texture, CubeMap


class Material:
    def __init__(self, name: str, vs: str, fs: str) -> None:
        self.name = name
        self.vs = vs
        self.fs = fs
        self.color_texture: Optional[Texture] = None
        self.cubemap: Optional[CubeMap] = None

    @staticmethod
    def from_assets(name: str) -> 'Material':
        import pkgutil
        vs = pkgutil.get_data('glglue', f'assets/{name}.vs')
        if not vs:
            raise Exception()
        fs = pkgutil.get_data('glglue', f'assets/{name}.fs')
        if not fs:
            raise Exception()
        return Material(name, vs.decode('utf-8'), fs.decode('utf-8'))



import ctypes
import math


class Float3(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_float),
        ('y', ctypes.c_float),
        ('z', ctypes.c_float),
    ]
    __match_args__ = ('x', 'y', 'z')

    def copy(self) -> 'Float3':
        dst = type(self)()
        ctypes.pointer(dst)[0] = self
        return dst

    def get_length(self) -> float:
        return math.sqrt(Float3.dot(self, self))

    def normalized(self) -> 'Float3':
        f = 1 / self.get_length()
        return self * f

    @staticmethod
    def dot(lhs, rhs) -> float:
        match lhs, rhs:
            case Float3(lx, ly, lz), Float3(rx, ry, rz):
                return lx*rx + ly*ry + lz*rz
        raise NotImplementedError()

    @staticmethod
    def cross(lhs, rhs):
        match lhs, rhs:
            case Float3(lx, ly, lz), Float3(rx, ry, rz):
                return Float3(
                    ly * rz - lz * ry,
                    lz * rx - lx * rz,
                    lx * ry - ly * rx)
        raise NotImplementedError()

    def __str__(self) -> str:
        return f'({self.x:.3f}, {self.y:.3f}, {self.z:.3f})'

    def __iter__(self):
        yield from (self.x, self.y, self.z)

    def __eq__(self, rhs):
        match rhs:
            case Float3(x, y, z) | (x, y, z):
                return self.x == x and self.y == y and self.z == z
        return False

    def __neg__(self) -> 'Float3':
        return Float3(-self.x, -self.y, -self.z)

    def __add__(self, rhs) -> 'Float3':
        match rhs:
            case Float3(x, y, z) | (x, y, z):
                return Float3(self.x + x, self.y + y, self.z + z)
        raise RuntimeError()

    def __sub__(self, rhs) -> 'Float3':
        match rhs:
            case Float3(x, y, z) | (x, y, z):
                return Float3(self.x - x, self.y - y, self.z - z)
        raise RuntimeError()

    def __mul__(self, rhs: float) -> 'Float3':
        return Float3(self.x * rhs, self.y * rhs, self.z * rhs)

    @staticmethod
    def new_inifinity() -> 'Float3':
        return Float3(float('inf'), float('inf'), float('inf'))



import logging

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(levelname)s:%(name)s:%(message)s',
                    level=logging.DEBUG)
import sys
from PyQt5 import Qt
from OpenGL.GL import *


class Controller:
    """
    [CLASSES] Controllerクラスは、glglueの規約に沿って以下のコールバックを実装する
    """
    def __init__(self):
        pass

    def onResize(self, w, h):
        logger.debug('onResize: %d, %d', w, h)
        glViewport(0, 0, w, h)

    def onLeftDown(self, x, y):
        logger.debug('onLeftDown: %d, %d', x, y)

    def onLeftUp(self, x, y):
        logger.debug('onLeftUp: %d, %d', x, y)

    def onMiddleDown(self, x, y):
        logger.debug('onMiddleDown: %d, %d', x, y)

    def onMiddleUp(self, x, y):
        logger.debug('onMiddleUp: %d, %d', x, y)

    def onRightDown(self, x, y):
        logger.debug('onRightDown: %d, %d', x, y)

    def onRightUp(self, x, y):
        logger.debug('onRightUp: %d, %d', x, y)

    def onMotion(self, x, y):
        logger.debug('onMotion: %d, %d', x, y)

    def onWheel(self, d):
        logger.debug('onWheel: %d', d)

    def onKeyDown(self, keycode):
        logger.debug('onKeyDown: %d', keycode)

    def onUpdate(self, d):
        #logger.debug('onUpdate: delta %d ms', d)
        pass

    def draw(self):
        glClearColor(0.0, 0.0, 1.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glBegin(GL_TRIANGLES)
        glVertex(-1.0, -1.0)
        glVertex(1.0, -1.0)
        glVertex(0.0, 1.0)
        glEnd()

        glFlush()


class Window(Qt.QMainWindow):
    def __init__(self, parent=None):
        import glglue.qgl5
        super().__init__(parent)
        # setup opengl widget
        self.controller = Controller()
        self.glwidget = glglue.qgl5.Widget(self, self.controller)
        self.setCentralWidget(self.glwidget)


if __name__ == "__main__":
    app = Qt.QApplication(sys.argv)
    window = Window()
    window.show()
    sys.exit(app.exec_())



from OpenGL import GL


class Cube(object):
    def __init__(self, s):
        self.vertices = [
            -s, -s, -s,
            s, -s, -s,
            s,  s, -s,
            -s,  s, -s,
            -s, -s,  s,
            s, -s,  s,
            s,  s,  s,
            -s,  s,  s,
        ]
        self.colors = [
            0, 0, 0,
            1, 0, 0,
            0, 1, 0,
            0, 0, 1,
            0, 1, 1,
            1, 0, 1,
            1, 1, 1,
            1, 1, 0,
        ]
        self.indices = [
            0, 1, 2, 2, 3, 0,
            0, 4, 5, 5, 1, 0,
            1, 5, 6, 6, 2, 1,
            2, 6, 7, 7, 3, 2,
            3, 7, 4, 4, 0, 3,
            4, 7, 6, 6, 5, 4,
        ]

    def draw(self):
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glVertexPointer(3, GL.GL_FLOAT, 0, self.vertices)
        GL.glEnableClientState(GL.GL_COLOR_ARRAY)
        GL.glColorPointer(3, GL.GL_FLOAT, 0, self.colors)
        GL.glDrawElements(GL.GL_TRIANGLES, len(self.indices),
                          GL.GL_UNSIGNED_INT, self.indices)
        GL.glDisableClientState(GL.GL_COLOR_ARRAY)
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)



from typing import Optional
import glfw
import logging
from OpenGL import GL
from .basecontroller import BaseController
from .windowconfig import WindowConfig

logger = logging.getLogger(__name__)


class LoopManager:
    def __init__(
            self, controller: BaseController, *,
            title: str = 'glfw',
            width: int = 0,
            height: int = 0,
            is_maximized: bool = False,
            config:  Optional[WindowConfig] = None
    ):
        width = width or (config.width if config else 1280)
        height = height or (config.height if config else 720)
        is_maximized = is_maximized or (
            config.is_maximized if config else False)

        self.controller = controller
        self.mouse_x = 0
        self.mouse_y = 0

        if not glfw.init():
            logger.error("Could not initialize OpenGL context")
            return

        # OS X supports only forward-compatible core profiles from 3.2
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 2)
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, GL.GL_TRUE)

        # Create a windowed mode window and its OpenGL context
        self.window = glfw.create_window(
            width, height, title, None, None
        )
        if not self.window:
            logger.error("Could not initialize Window")
            return

        if is_maximized:
            glfw.maximize_window(self.window)

        glfw.make_context_current(self.window)
        glfw.swap_interval(1)
        glfw.set_key_callback(self.window, self.keyboard_callback)
        glfw.set_cursor_pos_callback(self.window, self.cursor_calblack)
        glfw.set_mouse_button_callback(self.window, self.mouse_callback)
        glfw.set_window_size_callback(self.window, self.resize_callback)
        glfw.set_char_callback(self.window, self.char_callback)
        glfw.set_scroll_callback(self.window, self.scroll_callback)
        self.controller.onResize(width, height)

    def __del__(self):
        del self.controller
        glfw.terminate()

    def get_config(self) -> WindowConfig:
        w, h = glfw.get_window_size(self.window)
        x, y = glfw.get_window_pos(self.window)
        is_maximized = True if glfw.get_window_attrib(
            self.window, glfw.MAXIMIZED) else False
        return WindowConfig(x, y, w, h, is_maximized)

    def keyboard_callback(self, window, key, scancode, action, mods):
        pass
        # if action == glfw.PRESS:
        #     io.KeysDown[key] = True
        # elif action == glfw.RELEASE:
        #     io.KeysDown[key] = False

        # io.KeyCtrl = (
        #     io.KeysDown[glfw.KEY_LEFT_CONTROL] or
        #     io.KeysDown[glfw.KEY_RIGHT_CONTROL]
        # )

        # io.KeyAlt = (
        #     io.KeysDown[glfw.KEY_LEFT_ALT] or
        #     io.KeysDown[glfw.KEY_RIGHT_ALT]
        # )

        # io.KeyShift = (
        #     io.KeysDown[glfw.KEY_LEFT_SHIFT] or
        #     io.KeysDown[glfw.KEY_RIGHT_SHIFT]
        # )

        # io.KeySuper = (
        #     io.KeysDown[glfw.KEY_LEFT_SUPER] or
        #     io.KeysDown[glfw.KEY_RIGHT_SUPER]
        # )

    def char_callback(self, window, char):
        pass
        # if 0 < char < 0x10000:
        #     self.io.add_input_character(char)

    def resize_callback(self, window, width, height):
        self.controller.onResize(width, height)

    def cursor_calblack(self, window, x, y):
        self.controller.onMotion(x, y)
        self.mouse_x = x
        self.mouse_y = y

    def mouse_callback(self, window,  button, action, mods):
        match button, action:
            case glfw.MOUSE_BUTTON_LEFT, glfw.PRESS:
                self.controller.onLeftDown(self.mouse_x, self.mouse_y)
            case glfw.MOUSE_BUTTON_LEFT, glfw.RELEASE:
                self.controller.onLeftUp(self.mouse_x, self.mouse_y)
            case glfw.MOUSE_BUTTON_RIGHT, glfw.PRESS:
                self.controller.onRightDown(self.mouse_x, self.mouse_y)
            case glfw.MOUSE_BUTTON_RIGHT, glfw.RELEASE:
                self.controller.onRightUp(self.mouse_x, self.mouse_y)
            case glfw.MOUSE_BUTTON_MIDDLE, glfw.PRESS:
                self.controller.onMiddleDown(self.mouse_x, self.mouse_y)
            case glfw.MOUSE_BUTTON_MIDDLE, glfw.RELEASE:
                self.controller.onMiddleUp(self.mouse_x, self.mouse_y)

    def scroll_callback(self, window, x_offset, y_offset):
        self.controller.onWheel(y_offset)

    def begin_frame(self):
        if glfw.window_should_close(self.window):
            return None
        glfw.poll_events()
        return glfw.get_time() * 1000

    def end_frame(self):
        glfw.swap_buffers(self.window)



from typing import NamedTuple
from .float3 import Float3
from .mat4 import Mat4


class AABB(NamedTuple):
    min: Float3
    max: Float3

    def __str__(self) -> str:
        return f'AABB({self.min}, {self.max})'

    def expand(self, rhs: 'AABB') -> 'AABB':
        min = self.min.copy()
        if rhs.min.x < min.x:
            min.x = rhs.min.x
        if rhs.min.y < min.y:
            min.y = rhs.min.y
        if rhs.min.z < min.z:
            min.z = rhs.min.z

        max = self.max.copy()
        if rhs.max.x > max.x:
            max.x = rhs.max.x
        if rhs.max.y > max.y:
            max.y = rhs.max.y
        if rhs.max.z > max.z:
            max.z = rhs.max.z

        return AABB(min, max)

    def transform(self, m: Mat4) -> 'AABB':
        p0 = m.apply(*self.min)
        p1 = m.apply(*self.max)
        min_x, max_x = (p0.x, p1.x) if p0.x < p1.x else (p1.x, p0.x)
        min_y, max_y = (p0.y, p1.y) if p0.y < p1.y else (p1.y, p0.y)
        min_z, max_z = (p0.z, p1.z) if p0.z < p1.z else (p1.z, p0.z)
        return AABB(Float3(min_x, min_y, min_z), Float3(max_x, max_y, max_z))

    @staticmethod
    def new_empty() -> 'AABB':
        return AABB(Float3.new_inifinity(), -Float3.new_inifinity())



from typing import Optional
import ctypes
from OpenGL import GL
#
import pydear as ImGui
#
from ..ctypesmath import Camera, Float4
from .rendertarget import RenderTarget
from .samplecontroller import Scene


class RenderView:
    def __init__(self) -> None:
        self.camera = Camera()
        self.scene = Scene()
        self.render_target: Optional[RenderTarget] = None
        self.clear_color = (0.2, 0.2, 0.2, 1)
        self.light = Float4(1, 1, 1, 1)
        self.hovered = False

    def __del__(self):
        if self.render_target:
            del self.render_target
            self.render_target = None

    def _render(self, width: int, height: int) -> int:
        if self.render_target:
            if self.render_target.width != width or self.render_target.height != height:
                del self.render_target
                self.render_target = None
        if width <= 0 or height <= 0:
            return 0
        if not self.render_target:
            self.render_target = RenderTarget(width, height)

        #
        # update view camera
        #
        self.camera.onResize(width, height)

        #
        # render
        #
        with self.render_target.bind_context():
            GL.glViewport(0, 0, width, height)
            GL.glClearColor(*self.clear_color)
            GL.glClear(GL.GL_COLOR_BUFFER_BIT |
                       GL.GL_DEPTH_BUFFER_BIT)  # type: ignore

            state = self.camera.get_state(self.light)
            if self.scene:
                self.scene.draw(state)

        # <class 'numpy.uintc'>
        return int(self.render_target.texture)

    def draw(self, p_open: ctypes.Array):
        '''
        button に fbo を描画する
        '''
        ImGui.PushStyleVar_2(
            ImGui.ImGuiStyleVar_.WindowPadding, ImGui.ImVec2(0, 0))
        if ImGui.Begin(
                "render target", None, ImGui.ImGuiWindowFlags_.NoScrollbar | ImGui.ImGuiWindowFlags_.NoScrollWithMouse):
            w, h = ImGui.GetContentRegionAvail()
            x, y = ImGui.GetWindowPos()
            y += ImGui.GetFrameHeight()
            io = ImGui.GetIO()

            mouse_x = io.MousePos.x - x
            mouse_y = io.MousePos.y - y

            if ImGui.IsMouseDown(0):
                self.camera.onLeftDown(mouse_x, mouse_y)
            elif ImGui.IsMouseReleased(0):
                self.camera.onLeftUp(mouse_x, mouse_y)
            if ImGui.IsMouseDown(1):
                self.camera.onRightDown(mouse_x, mouse_y)
            elif ImGui.IsMouseReleased(1):
                self.camera.onRightUp(mouse_x, mouse_y)
            if ImGui.IsMouseDown(2):
                self.camera.onMiddleDown(mouse_x, mouse_y)
            elif ImGui.IsMouseReleased(2):
                self.camera.onMiddleUp(mouse_x, mouse_y)

            if self.hovered and io.MouseWheel:
                self.camera.onWheel(-int(io.MouseWheel))

            if ImGui.IsMouseDragging(0) or ImGui.IsMouseDragging(1) or ImGui.IsMouseDragging(2):
                self.camera.onMotion(mouse_x, mouse_y)

            texture = self._render(int(w), int(h))
            if texture:
                # ImGui.ImageButton(
                #     ctypes.c_void_p(texture), (w, h), (0.0, 0.0), (1.0, 1.0), 0, bg_col=ImGui.ImVec4(0, 0, 0, 1), tint_col=ImGui.ImVec4(1, 1, 1, 1))
                # https://gamedev.stackexchange.com/questions/140693/how-can-i-render-an-opengl-scene-into-an-imgui-window
                # Using a Child allow to fill all the space of the window.
                # It also alows customization
                ImGui.BeginChild(b"cameraview")
                # Because I use the texture from OpenGL, I need to invert the V from the UV.
                ImGui.Image(ctypes.c_void_p(texture), (w, h), (0, 1), (1, 0))
                self.hovered = ImGui.IsItemHovered()
                ImGui.EndChild()
        ImGui.End()
        ImGui.PopStyleVar()



#
# pip install pyside6
#
from PySide6 import QtWidgets, QtGui
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(levelname)s:%(name)s:%(message)s',
                    level=logging.DEBUG)


class Window(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        # setup opengl widget
        import glglue.gl3.samplecontroller
        self.controller = glglue.gl3.samplecontroller.SampleController()
        import glglue.pyside6
        import glglue.utils
        self.glwidget = glglue.pyside6.Widget(
            self, self.controller, glglue.utils.get_desktop_scaling_factor())
        self.setCentralWidget(self.glwidget)


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    window.show()
    sys.exit(app.exec())



import contextlib
from OpenGL import GL


class RenderTarget:
    def __init__(self, width: int, height: int) -> None:
        self.fbo = GL.glGenFramebuffers(1)
        self.width = width
        self.height = height
        # color
        self.texture = GL.glGenTextures(1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, width,
                        height, 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, None)
        GL.glTexParameteri(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexParameteri(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        # depth
        self.depth = GL.glGenRenderbuffers(1)
        GL.glBindRenderbuffer(GL.GL_RENDERBUFFER, self.depth)
        GL.glRenderbufferStorage(
            GL.GL_RENDERBUFFER, GL.GL_DEPTH_COMPONENT, width, height)

        # bind fbo with texture nd depth
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fbo)
        GL.glFramebufferTexture2D(
            GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0, GL.GL_TEXTURE_2D, self.texture, 0)
        GL.glFramebufferRenderbuffer(
            GL.GL_FRAMEBUFFER, GL.GL_DEPTH_ATTACHMENT, GL.GL_RENDERBUFFER, self.depth)
        GL.glDrawBuffers([GL.GL_COLOR_ATTACHMENT0])
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)

        # logger.debug(
        #     f'create fbo {self.fbo}, texture {self.texture}, {self.depth}: {width} x {height}')

    def __del__(self):
        # logger.debug(
        #     f'delte fbo {self.fbo}, texture {self.texture}, {self.depth}')
        if self.fbo:
            GL.glDeleteFramebuffers(1, [self.fbo])
            self.fbo = 0
        if self.texture:
            GL.glDeleteTextures([self.texture])
            self.texture = 0
        if self.depth:
            GL.glDeleteRenderbuffers(1, [self.depth])
            self.depth = 0

    def bind(self):
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fbo)

    def unbind(self):
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)

    @contextlib.contextmanager
    def bind_context(self):
        self.bind()
        try:
            yield
        finally:
            self.unbind()



import unittest
import glglue.utils


class TestUtils(unittest.TestCase):

    def test_utils(self):
        self.assertEqual(1.5, glglue.utils.get_desktop_scaling_factor())


if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod
from logging import getLogger
from typing import Any, List
from OpenGL import GL
from . import gizmo
import glglue.basecontroller
from glglue.ctypesmath import Camera, FrameState, AABB, Float4
import glglue.gl3.vbo
import glglue.scene.material
from . renderer import Renderer
logger = getLogger(__name__)


class Scene:
    def __init__(self) -> None:
        self.env: List[Any] = []
        from ..scene import cube
        self.drawables: List[Any] = [cube.create_cube(0.3)]
        self.renderer = Renderer()
        self.gizmo = gizmo.Gizmo()

    def __del__(self):
        del self.renderer

    def update(self, d: int) -> bool:
        updated = False
        for drawable in self.drawables:
            if drawable.update(d):
                updated = True
        return updated

    def draw(self, state: FrameState):
        self.gizmo.begin(state)
        self.gizmo.axis(10)

        for e in self.env:
            self.renderer.draw(e, state)
        for _, drawable in enumerate(self.drawables):
            self.renderer.draw(drawable, state)
            # aabb
            aabb = AABB.new_empty()
            self.gizmo.aabb(drawable.expand_aabb(aabb))

        self.gizmo.end()


class SampleController(glglue.basecontroller.BaseController):
    def __init__(self):
        self.clear_color = (0.6, 0.6, 0.4, 0.0)
        self.camera = Camera()
        self.scene = Scene()
        self.isInitialized = False
        self.light = Float4(1, 1, 1, 1)

    def onResize(self, w: int, h: int) -> bool:
        return self.camera.onResize(w, h)

    def onLeftDown(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onLeftDown(x, y)

    def onLeftUp(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onLeftUp(x, y)

    def onMiddleDown(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onMiddleDown(x, y)

    def onMiddleUp(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onMiddleUp(x, y)

    def onRightDown(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onRightDown(x, y)

    def onRightUp(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onRightUp(x, y)

    def onMotion(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onMotion(x, y)

    def onWheel(self, d: int) -> bool:
        ''' mouse input '''
        return self.camera.onWheel(d)

    def onKeyDown(self, keycode: int) -> bool:
        return False

    def onUpdate(self, d: int) -> bool:
        '''
        milliseconds
        '''
        if not self.scene:
            return False
        return self.scene.update(d)

    def initialize(self):
        import glglue
        logger.info(glglue.get_info())
        self.isInitialized = True

    def draw(self):
        if not self.isInitialized:
            self.initialize()
        GL.glClearColor(*self.clear_color)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT |
                   GL.GL_DEPTH_BUFFER_BIT)  # type: ignore

        state = self.camera.get_state(self.light)
        GL.glViewport(int(state.viewport.x), int(state.viewport.y),
                      int(state.viewport.z), int(state.viewport.w))

        if self.scene:
            self.scene.draw(state)

        GL.glFlush()



import ctypes
import array
from OpenGL import GL
#
from .node import Node
from .mesh import Mesh
from .skin import Skin
from .material import Material
from .vertices import Interleaved, VectorView
from glglue.ctypesmath import Float3, Mat4


class Vertex(ctypes.Structure):
    _fields_ = [
        ('position', Float3),
        ('normal', Float3),
    ]


def create_skin() -> Node:
    green = Float3(0, 1, 0)
    red = Float3(1, 0, 0)
    s = 0.1
    h1 = s*2
    h2 = s*4
    h3 = s*6
    vertices = (Vertex * 16)(
        Vertex(Float3(-s, 0, s), red),
        Vertex(Float3(s, 0, s), red),
        Vertex(Float3(s, 0, -s), red),
        Vertex(Float3(-s, 0, -s), red),
        Vertex(Float3(-s, h1, s), red),
        Vertex(Float3(s, h1, s), red),
        Vertex(Float3(s, h1, -s), red),
        Vertex(Float3(-s, h1, -s), red),
        Vertex(Float3(-s, h2, s), green),
        Vertex(Float3(s, h2, s), green),
        Vertex(Float3(s, h2, -s), green),
        Vertex(Float3(-s, h2, -s), green),
        Vertex(Float3(-s, h3, s), green),
        Vertex(Float3(s, h3, s), green),
        Vertex(Float3(s, h3, -s), green),
        Vertex(Float3(-s, h3, -s), green),
    )
    indices = array.array('H')

    def push_quad(i0, i1, i2, i3):
        indices.append(i0)
        indices.append(i1)
        indices.append(i2)
        #
        indices.append(i2)
        indices.append(i3)
        indices.append(i0)
    '''
      +---+
     /   /|  15 14
    +---+ +12 13
    |   |/|  11 10
    +---+ + 8  9
    |   |/|   7  6
    +---+ + 4  5
    |   |/    3  2
    +---+   0  1
    '''
    push_quad(0, 4, 5, 1)
    push_quad(4, 8, 9, 5)
    push_quad(8, 12, 13, 9)

    push_quad(1, 5, 6, 2)
    push_quad(5, 9, 10, 6)
    push_quad(9, 13, 14, 10)

    push_quad(2, 6, 7, 3)
    push_quad(6, 10, 11, 7)
    push_quad(10, 14, 15, 11)

    push_quad(3, 7, 4, 0)
    push_quad(7, 11, 8, 4)
    push_quad(11, 15, 12, 8)

    mesh = Mesh(
        'skin',
        Interleaved(VectorView(memoryview(vertices),
                    ctypes.c_float, 6), [0, 12]),
        VectorView.create(indices))
    material = Material.from_assets('cube')

    mesh.add_submesh(material, [], GL.GL_TRIANGLES)

    bone0 = Node('bone0', Mat4.new_identity())
    bone1 = Node('bone1', Mat4.new_translation(0, h2, 0))
    bone0.children.append(bone1)
    bone0.meshes.append(mesh)

    # bone0.skin = Skin([bone0, bone1])

    return bone0



# coding: utf-8
__version__ = "1.2.0"
from .basecontroller import BaseController
from . import gl3


def get_info():
    from OpenGL.GL import (glGetString, glGetIntegerv, GL_VENDOR, GL_VERSION,
                           GL_SHADING_LANGUAGE_VERSION, GL_RENDERER,
                           GL_MAJOR_VERSION, GL_MINOR_VERSION)
    return {
        'major': glGetIntegerv(GL_MAJOR_VERSION),
        'minor': glGetIntegerv(GL_MINOR_VERSION),
        'vendor': glGetString(GL_VENDOR),
        'version': glGetString(GL_VERSION),
        'shader_language_version': glGetString(GL_SHADING_LANGUAGE_VERSION),
        'renderer': glGetString(GL_RENDERER),
    }



from typing import NamedTuple, List, Type
import array
import ctypes


MAP = {
    'B': ctypes.c_uint8,
    'H': ctypes.c_int16,
    'I': ctypes.c_uint32,
}


class VectorView(NamedTuple):
    # only use memoryview.nbytes
    data: memoryview
    # base type
    element_type: Type[ctypes._SimpleCData]
    # vector element count. vec2, vec3, mat4...etc
    element_count: int = 1

    @staticmethod
    def create(src: array.array) -> 'VectorView':
        match src:
            case array.array() as a:
                return VectorView(memoryview(a), MAP[a.typecode])
        raise RuntimeError()

    def get_stride(self) -> int:
        return ctypes.sizeof(self.element_type) * self.element_count

    def get_count(self) -> int:
        return self.data.nbytes // self.get_stride()


class Planar(NamedTuple):
    attributes: List[VectorView]

    def count(self) -> int:
        return self.attributes[0].get_count()


class Interleaved(NamedTuple):
    vertices: VectorView
    offsets: List[int]

    def count(self) -> int:
        return self.vertices.get_count()



from typing import Dict, Union, Tuple


from ..scene.node import Node
from ..scene.mesh import Mesh
from ..scene.material import Material, Texture, CubeMap
from ..ctypesmath import Mat4, FrameState, Float4
import glglue.gl3.vbo
import glglue.gl3.shader
import glglue.gl3.texture
from OpenGL import GL


class Renderer:
    def __init__(self) -> None:
        self.textures: Dict[Texture, glglue.gl3.texture.Texture] = {}
        self.cubemaps: Dict[CubeMap, glglue.gl3.texture.CubeMap] = {}
        self.shaders: Dict[glglue.gl3.shader.ShaderSource,
                           glglue.gl3.shader.Shader] = {}
        self.meshes: Dict[Mesh, glglue.gl3.vbo.Drawable] = {}

    def _get_or_create_texture(self, src: Texture) -> glglue.gl3.texture.Texture:
        texture = self.textures.get(src)
        if not texture:
            texture = glglue.gl3.texture.Texture()
            image = src.image
            texture.load(image.data, image.width, image.height)
            self.textures[src] = texture
        return texture

    def _get_or_create_cubemap(self, src: CubeMap) -> glglue.gl3.texture.CubeMap:
        cubemap = self.cubemaps.get(src)
        if not cubemap:
            cubemap = glglue.gl3.texture.CubeMap()
            cubemap.load(*src)
            self.cubemaps[src] = cubemap
        return cubemap

    def _get_or_create_shader(self, src: Material, macro: Tuple[str, ...]) -> glglue.gl3.shader.Shader:
        shader_source = glglue.gl3.shader.ShaderSource(
            src.vs, macro, src.fs, macro)
        shader = self.shaders.get(shader_source)
        if not shader:
            shader = glglue.gl3.shader.create_from(shader_source)
            self.shaders[shader_source] = shader
        return shader

    def _get_or_create_mesh(self, src: Mesh) -> glglue.gl3.vbo.Drawable:
        drawable = self.meshes.get(src)
        if not drawable:
            drawable = glglue.gl3.vbo.create(src.vertices, src.indices)
            # dm = DrawableMacro(mesh, macro)
            self.meshes[src] = drawable
        return drawable

    def _draw_mesh(self, mesh: Mesh, projection: Mat4, view: Mat4, model: Mat4, light: Float4):
        drawable = self._get_or_create_mesh(mesh)

        for submesh in mesh.submeshes:

            shader = self._get_or_create_shader(
                submesh.material, tuple(submesh.macro))

            shader.use()

            material = submesh.material
            if material.cubemap:

                shader.set_uniform('V', view, False)
                shader.set_uniform('P', projection, False)

                GL.glEnable(GL.GL_CULL_FACE)
                GL.glDepthMask(GL.GL_FALSE)
                cubemap = self._get_or_create_cubemap(material.cubemap)
                cubemap.bind()
                GL.glActiveTexture(GL.GL_TEXTURE0)
                drawable.draw(submesh.topology, submesh.offset,
                              submesh.draw_count)
                GL.glDepthMask(GL.GL_TRUE)

            else:

                shader.set_uniform('vp', view * projection)
                shader.set_uniform('m', model)
                shader.set_uniform('v', view)
                shader.set_uniform('p', projection)
                shader.set_uniform('light', light)

                if submesh.material.color_texture:
                    texture = self._get_or_create_texture(
                        submesh.material.color_texture)
                    shader.set_texture('COLOR_TEXTURE', 0, texture)

                drawable.draw(submesh.topology, submesh.offset,
                              submesh.draw_count)

    def _draw_node(self, node: Node, projection: Mat4, view: Mat4, parent: Mat4, light: Float4):
        m = node.get_local_matrix() * parent

        for mesh in node.meshes:
            self._draw_mesh(mesh, projection, view, m, light)

        for child in node.children:
            self._draw_node(child, projection, view, m, light)

    def draw(self, root: Union[Node, Mesh], state: FrameState):
        match root:
            case Node() as node:
                self._draw_node(node, state.camera_projection,
                                state.camera_view, Mat4.new_identity(), state.light)
            case Mesh() as mesh:
                self._draw_mesh(mesh, state.camera_projection,
                                state.camera_view, Mat4.new_identity(), state.light)



import ctypes
import math
from typing import Tuple

from glglue.ctypesmath.float3 import Float3


def dot(l, r):
    value = 0
    for ll, rr in zip(l, r):
        value += ll * rr
    return value


class Mat4(ctypes.Structure):
    _fields_ = [("_11", ctypes.c_float), ("_12", ctypes.c_float),
                ("_13", ctypes.c_float), ("_14", ctypes.c_float),
                ("_21", ctypes.c_float), ("_22", ctypes.c_float),
                ("_23", ctypes.c_float), ("_24", ctypes.c_float),
                ("_31", ctypes.c_float), ("_32", ctypes.c_float),
                ("_33", ctypes.c_float), ("_34", ctypes.c_float),
                ("_41", ctypes.c_float), ("_42", ctypes.c_float),
                ("_43", ctypes.c_float), ("_44", ctypes.c_float)]

    def __str__(self) -> str:
        return f'[{self._11}, {self._12}, {self._13}, {self._14}]' + f'[{self._21}, {self._22}, {self._23}, {self._24}]' + f'[{self._31}, {self._32}, {self._33}, {self._34}]' + f'[{self._41}, {self._42}, {self._43}, {self._44}]'

    def __eq__(self, rhs: 'Mat4') -> bool:
        return (self._11 == rhs._11 and self._12 == rhs._12 and self._13 == rhs._13 and self._14 == rhs._14
                and self._21 == rhs._21 and self._22 == rhs._22 and self._23 == rhs._23 and self._24 == rhs._24
                and self._31 == rhs._31 and self._32 == rhs._32 and self._33 == rhs._33 and self._34 == rhs._34
                and self._41 == rhs._41 and self._42 == rhs._42 and self._43 == rhs._43 and self._44 == rhs._44
                )

    def __mul__(self, rhs: 'Mat4') -> 'Mat4':
        m = Mat4()

        def row(row):
            match row:
                case 1:
                    return self._11, self._12, self._13, self._14
                case 2:
                    return self._21, self._22, self._23, self._24
                case 3:
                    return self._31, self._32, self._33, self._34
                case 4:
                    return self._41, self._42, self._43, self._44
                case _:
                    raise RuntimeError()

        def col(col):
            match col:
                case 1:
                    return rhs._11, rhs._21, rhs._31, rhs._41
                case 2:
                    return rhs._12, rhs._22, rhs._32, rhs._42
                case 3:
                    return rhs._13, rhs._23, rhs._33, rhs._43
                case 4:
                    return rhs._14, rhs._24, rhs._34, rhs._44
                case _:
                    return RuntimeError()

        m._11 = dot(row(1), col(1))
        m._12 = dot(row(1), col(2))
        m._13 = dot(row(1), col(3))
        m._14 = dot(row(1), col(4))

        m._21 = dot(row(2), col(1))
        m._22 = dot(row(2), col(2))
        m._23 = dot(row(2), col(3))
        m._24 = dot(row(2), col(4))

        m._31 = dot(row(3), col(1))
        m._32 = dot(row(3), col(2))
        m._33 = dot(row(3), col(3))
        m._34 = dot(row(3), col(4))

        m._41 = dot(row(4), col(1))
        m._42 = dot(row(4), col(2))
        m._43 = dot(row(4), col(3))
        m._44 = dot(row(4), col(4))

        return m

    def transposed(self, *, no_translation=False) -> 'Mat4':
        if no_translation:
            return Mat4(
                self._11, self._21, self._31, 0,
                self._12, self._22, self._32, 0,
                self._13, self._23, self._33, 0,
                0, 0, 0, 1
            )
        else:
            return Mat4(
                self._11, self._21, self._31, self._41,
                self._12, self._22, self._32, self._42,
                self._13, self._23, self._33, self._43,
                self._14, self._24, self._34, self._44
            )

    def apply(self, x: float, y: float, z: float, *, translate=True) -> Float3:
        if translate:
            return Float3(
                x * self._11 + y * self._21 + z * self._31 + self._41,
                x * self._12 + y * self._22 + z * self._32 + self._42,
                x * self._13 + y * self._23 + z * self._33 + self._43
            )
        else:
            return Float3(
                x * self._11 + y * self._21 + z * self._31,
                x * self._12 + y * self._22 + z * self._32,
                x * self._13 + y * self._23 + z * self._33
            )

    def to_array(self):
        return (ctypes.c_float * 16).from_buffer(self)

    def perspective(self, fov_y: float, aspect: float, z_near: float,
                    z_far: float) -> None:
        fov_y *= 0.5
        cot = 1.0 / math.tan(fov_y)
        self._11 = cot / aspect
        self._12 = 0
        self._13 = 0
        self._14 = 0
        self._21 = 0
        self._22 = cot
        self._23 = 0
        self._24 = 0
        self._31 = 0
        self._32 = 0
        self._33 = -(z_far + z_near) / (z_far - z_near)
        self._34 = -1
        self._41 = 0
        self._42 = 0
        self._43 = -2 * z_far * z_near / (z_far - z_near)
        self._44 = 0

    def inverse_rigidbody(self) -> 'Mat4':
        '''
        inverse only rotation + translation. no scale

        (R x T)^-1 = T^-1 * R^-1
        '''
        return Mat4.new_translation(-self._41, -self._42, -self._43) * self.transposed(no_translation=True)

    @classmethod
    def new_perspective(cls, fov_y, aspect, z_near, z_far) -> 'Mat4':
        m = cls()
        m.perspective(fov_y, aspect, z_near, z_far)
        return m

    @classmethod
    def new_identity(cls):
        return cls(
            1.0,
            0.0,
            0.0,
            0.0,  #
            0.0,
            1.0,
            0.0,
            0.0,  #
            0.0,
            0.0,
            1.0,
            0.0,  #
            0.0,
            0.0,
            0.0,
            1.0  #
        )

    @staticmethod
    def new_scale(x: float, y: float, z: float) -> 'Mat4':
        return Mat4(
            x, 0, 0, 0,
            0, y, 0, 0,
            0, 0, z, 0,
            0, 0, 0, 1,
        )

    @staticmethod
    def new_translation(x: float, y: float, z: float) -> 'Mat4':
        return Mat4(
            1, 0, 0, 0,  #
            0, 1, 0, 0,  #
            0, 0, 1, 0,  #
            x, y, z, 1  #
        )

    @classmethod
    def new_rotation_z(cls, rad):
        s = math.sin(rad)
        c = math.cos(rad)
        return cls(
            c, s, 0, 0,  #
            -s, c, 0, 0,  #
            0, 0, 1, 0,  #
            0, 0, 0, 1  #
        )

    @classmethod
    def new_rotation_y(cls, rad):
        s = math.sin(rad)
        c = math.cos(rad)
        return cls(
            c, 0, -s, 0,  #
            0, 1, 0, 0,  #
            s, 0, c, 0,  #
            0, 0, 0, 1  #
        )

    @classmethod
    def new_rotation_x(cls, rad):
        s = math.sin(rad)
        c = math.cos(rad)
        return cls(
            1, 0, 0, 0,  #
            0, c, s, 0,  #
            0, -s, c, 0,  #
            0, 0, 0, 1  #
        )

    @staticmethod
    def new_from_quaternion(x: float, y: float, z: float, w: float) -> 'Mat4':
        xx = x * x
        yy = y * y
        zz = z * z
        ww = w * w

        xy = x * y
        yz = y * z
        zx = z * x

        wx = w * x
        wy = w * y
        wz = w * z

        return Mat4(
            xx-yy-zz+ww, 2*xy+2*wz, 2*zx-2*wy, 0,
            2*xy-2*wz, -xx+yy-zz+ww, 2*yz+2*wx, 0,
            2*zx+2*wy, 2*yz-2*wx, -xx-yy+zz+ww, 0,
            0, 0, 0, 1
        )

    @staticmethod
    def new_axis_angle(axis, angle) -> 'Mat4':
        if angle == 0:
            return Mat4.new_identity()

        c = math.cos(angle)
        s = math.sin(angle)
        xx = axis.x * axis.x
        yy = axis.y * axis.y
        zz = axis.z * axis.z
        xy = axis.x * axis.y
        yz = axis.y * axis.z
        zx = axis.z * axis.x
        xs = axis.x * s
        ys = axis.y * s
        zs = axis.z * s
        return Mat4(
            c + xx * (1-c),  xy * (1-c) - zs, zx * (1-c) + ys, 0,
            xy * (1-c) + zs, c + yy * (1-c), yz * (1-c) - xs, 0,
            zx * (1-c) - ys, yz * (1-c) + xs, c + zz*(1-c), 0,
            0, 0, 0, 1
        )

    @staticmethod
    def new_coords(x: Float3, y: Float3, z: Float3, t: Float3):
        return Mat4(
            x.x, x.y, x.z, 0,
            y.x, y.y, y.z, 0,
            z.x, z.y, z.z, 0,
            t.x, t.y, t.z, 1,
        )


class Float4(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float), ("y", ctypes.c_float),
                ("z", ctypes.c_float), ("w", ctypes.c_float)]

    def __mul__(self, m: Mat4) -> 'Float4':
        v = Float4()
        v.x = self.x * m._11 + self.y * m._21 + self.z * m._31 + self.w * m._41
        v.y = self.x * m._12 + self.y * m._22 + self.z * m._32 + self.w * m._42
        v.z = self.x * m._13 + self.y * m._23 + self.z * m._33 + self.w * m._43
        v.w = self.x * m._14 + self.y * m._24 + self.z * m._34 + self.w * m._44
        return v



#
# pip install pyside6
#
from PySide6 import QtWidgets, QtGui, QtCore
import logging
from glglue.ctypesmath.camera import FrameState
from glglue.gl3.samplecontroller import Scene
from glglue.gl3 import gizmo
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(levelname)s:%(name)s:%(message)s',
                    level=logging.DEBUG)


class Scene(Scene):
    def __init__(self) -> None:
        super().__init__()
        self.gizmo = gizmo.Gizmo()

    def update(self, d: int):
        return True

    def draw(self, state: FrameState):
        self.gizmo.begin(state)
        self.gizmo.axis(10)

        if self.gizmo.bone('bone1', 1):
            logger.info('bone clicked')

        self.gizmo.end()


class Window(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        # setup opengl widget
        self.setWindowTitle('gizmo_sample')
        import glglue.gl3.samplecontroller
        self.controller = glglue.gl3.samplecontroller.SampleController()
        self.scene = Scene()
        self.controller.scene = self.scene  # type: ignore
        import glglue.pyside6
        import glglue.utils
        self.glwidget = glglue.pyside6.Widget(
            self, self.controller, glglue.utils.get_desktop_scaling_factor())
        self.setCentralWidget(self.glwidget)

        # render loop
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.glwidget.update)
        self.timer.start(33)


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    window.show()
    sys.exit(app.exec())



from typing import NamedTuple, Optional
import math
from .float3 import Float3

kEpsilon = 1e-5


class Ray(NamedTuple):
    origin: Float3
    dir: Float3

    def intersect(self, v0: Float3, v1: Float3, v2: Float3) -> Optional[float]:
        '''
        https://www.scratchapixel.com/lessons/3d-basic-rendering/ray-tracing-rendering-a-triangle/ray-triangle-intersection-geometric-solution
        '''
        # compute plane's normal
        v0v1 = v1 - v0
        v0v2 = v2 - v0
        # no need to normalize
        N = Float3.cross(v0v1, v0v2)  # N
        # area2 = N.get_length()

        # Step 1: finding P

        # check if ray and plane are parallel ?
        NdotRayDirection = Float3.dot(N, self.dir)
        if math.fabs(NdotRayDirection) < kEpsilon:  # almost 0
            return  # they are parallel so they don't intersect !

        # compute d parameter using equation 2
        d = -Float3.dot(N, v0)

        # compute t (equation 3)
        t = -(Float3.dot(N, self.origin) + d) / NdotRayDirection
        # check if the triangle is in behind the ray
        if t < 0:
            return  # the triangle is behind

        # compute the intersection point using equation 1
        P = self.origin + self.dir * t

        # Step 2: inside-outside test
        # Vec3f C  # vector perpendicular to triangle's plane

        # edge 0
        edge0 = v1 - v0
        vp0 = P - v0
        C = Float3.cross(edge0, vp0)
        if Float3.dot(N, C) < 0:
            return  # P is on the right side

        # edge 1
        edge1 = v2 - v1
        vp1 = P - v1
        C = Float3.cross(edge1, vp1)
        if Float3.dot(N, C) < 0:
            return  # P is on the right side

        # edge 2
        edge2 = v0 - v2
        vp2 = P - v2
        C = Float3.cross(edge2, vp2)
        if Float3.dot(N, C) < 0:
            return  # P is on the right side

        return t  # this ray hits the triangle



'''
https://antongerdelan.net/opengl/cubemaps.html
'''
import array
import pkgutil
from OpenGL import GL
from .vertices import Planar, VectorView
from .mesh import Mesh
from .material import Material
from .texture import CubeMap


def create_cubemap(cubemap: CubeMap, size: float = 10.0):
    vertices = array.array('f', [
        -size,  size, -size,
        -size, -size, -size,
        size, -size, -size,
        size, -size, -size,
        size,  size, -size,
        -size,  size, -size,

        -size, -size,  size,
        -size, -size, -size,
        -size,  size, -size,
        -size,  size, -size,
        -size,  size,  size,
        -size, -size,  size,

        size, -size, -size,
        size, -size,  size,
        size,  size,  size,
        size,  size,  size,
        size,  size, -size,
        size, -size, -size,

        -size, -size,  size,
        -size,  size,  size,
        size,  size,  size,
        size,  size,  size,
        size, -size,  size,
        -size, -size,  size,

        -size,  size, -size,
        size,  size, -size,
        size,  size,  size,
        size,  size,  size,
        -size,  size,  size,
        -size,  size, -size,

        -size, -size, -size,
        -size, -size,  size,
        size, -size, -size,
        size, -size, -size,
        -size, -size,  size,
        size, -size,  size
    ])

    mesh = Mesh(f'cubemap', Planar(
        [VectorView(memoryview(vertices).cast('f'), 3)]))

    vs = pkgutil.get_data('glglue', 'assets/cubemap.vs')
    if not vs:
        raise Exception()
    fs = pkgutil.get_data('glglue', 'assets/cubemap.fs')
    if not fs:
        raise Exception()
    material = Material('cubemap', vs.decode('utf-8'), fs.decode('utf-8'))
    material.cubemap = cubemap
    mesh.add_submesh(material, [], GL.GL_TRIANGLES)
    return mesh



import math
import ctypes
import array
from OpenGL import GL
from .mesh import Mesh
from .material import Material
from .vertices import Interleaved, VectorView


class Vertex(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_float),
        ('y', ctypes.c_float),
        ('z', ctypes.c_float),
        ('r', ctypes.c_float),
        ('g', ctypes.c_float),
        ('b', ctypes.c_float),
    ]


def to_radian(degree):
    return degree / 180.0 * math.pi

# VELOCITY = 0.1
# def rotator():
#     x_rot = 0.0
#     y_rot = 0.0
#     def update(self, delta_ms):
#         y_rot += delta_ms * VELOCITY
#         while y_rot > 360.0:
#             y_rot -= 360.0
#         x_rot += delta_ms * VELOCITY * 0.5
#         while x_rot > 360.0:
#             x_rot -= 360.0
#         self.m = glglue.ctypesmath.Mat4.new_rotation_y(to_radian(
#             self.y_rot)) * glglue.ctypesmath.Mat4.new_rotation_x(
#                 to_radian(self.x_rot))
#     return update


def create_cube(s: float) -> Mesh:
    vertices = (Vertex * 8)(
        Vertex(-s,
               -s,
               -s, 0, 0, 0),
        Vertex(s,
               -s,
               -s, 1, 0, 0),
        Vertex(s,
               s,
               -s, 0, 1, 0),
        Vertex(-s,
               s,
               -s, 0, 0, 1),
        Vertex(-s,
               -s,
               s, 0, 1, 1),
        Vertex(s,
               -s,
               s, 1, 0, 1),
        Vertex(s,
               s,
               s, 1, 1, 1),
        Vertex(-s,
               s,
               s, 1, 1, 0),
    )
    indices = array.array('H', [
        0, 1, 2,
        2, 3, 0,
        0, 4, 5,
        5, 1, 0,
        1, 5, 6,
        6, 2, 1,
        2, 6, 7,
        7, 3, 2,
        3, 7, 4,
        4, 0, 3,
        4, 7, 6,
        6, 5, 4,
    ])

    mesh = Mesh(f'cube {s}', Interleaved(
        VectorView(memoryview(vertices), ctypes.c_float, 6), [0, 12]),
        VectorView.create(indices))

    material = Material.from_assets('cube')

    mesh.add_submesh(material, [], GL.GL_TRIANGLES)
    return mesh



# coding: utf-8
'''
# glut
require pyOpenGL + glut.dll
require pyOpenGL + freeglut.dll

# glut install on Windows

* glut.dllもしくはfreeglut.dllを入手(vcpkgでビルドするのおすすめ)
* Python.exeと同じフォルダにコピーするか環境変数PATHを設定する
'''

import pathlib
import sys
HERE = pathlib.Path(__file__).absolute().parent
sys.path.insert(0, str(HERE.parent / 'src'))

if __name__ == "__main__":
    import glglue.sample
    import glglue.glut
    controller = glglue.sample.SampleController()

    # manual loop
    lastclock = 0
    loop = glglue.glut.LoopManager(controller)
    while True:
        # require freeglut glutMainLoopEvent
        clock = loop.begin_frame()
        d = clock - lastclock
        lastclock = clock
        controller.onUpdate(d)
        controller.draw()
        loop.end_frame()
 


import ctypes
import math
import logging
from typing import NamedTuple
from .mat4 import Mat4, Float4
from .float3 import Float3
from .hittest import Ray
logger = logging.getLogger(__name__)


class Perspective:
    def __init__(self) -> None:
        self.matrix = Mat4.new_identity()
        self.fov_y = math.pi * 30 / 180
        self.aspect = 1.0
        self.z_near = 0.1
        self.z_far = 1000
        self.update_matrix()

    def update_matrix(self) -> None:
        self.matrix.perspective(self.fov_y, self.aspect, self.z_near,
                                self.z_far)


class Orbit:
    def __init__(self) -> None:
        self.matrix = Mat4.new_identity()
        self.inverse = Mat4.new_identity()
        self.x = 0.0
        self.y = 0.0
        self.distance = 2.0
        self.yaw = 0.0
        self.pitch = 0.0
        self.update_matrix()

    def __str__(self) -> str:
        return f'({self.x:.3f}, {self.y:.3f}, {self.distance:.3f})'

    def update_matrix(self) -> None:
        t = Mat4.new_translation(self.x, self.y, -self.distance)
        yaw = Mat4.new_rotation_y(self.yaw)
        pitch = Mat4.new_rotation_x(self.pitch)
        self.matrix = yaw * pitch * t
        self.inverse = self.matrix.inverse_rigidbody()


class FrameState(NamedTuple):
    '''
    RenderInfo for frame
    '''
    viewport: Float4
    mouse_x: int
    mouse_y: int
    mouse_left_down: bool
    mouse_right_down: bool
    mouse_middle_down: bool
    camera_view: Mat4
    camera_projection: Mat4
    ray: Ray
    light: Float4


class Camera:
    def __init__(self):
        self.projection = Perspective()
        self.view = Orbit()
        self.width = 1
        self.height = 1
        self.x = 0
        self.y = 0
        self.left = False
        self.middle = False
        self.right = False

    def onResize(self, w: int, h: int) -> bool:
        if self.width == w and self.height == h:
            return False
        self.width = w
        self.height = h
        self.projection.aspect = float(w) / h
        self.projection.update_matrix()
        return True

    def onLeftDown(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        if self.left:
            return False
        self.left = True
        self.x = x
        self.y = y
        return False

    def onLeftUp(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        self.left = False
        return False

    def onMiddleDown(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        if self.middle:
            return False
        self.middle = True
        self.x = x
        self.y = y
        return False

    def onMiddleUp(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        self.middle = False
        return False

    def onRightDown(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        if self.right:
            return False
        self.right = True
        self.x = x
        self.y = y
        return False

    def onRightUp(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        self.right = False
        return False

    def onMotion(self, x: int, y: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        dx = x - self.x
        self.x = x
        dy = y - self.y
        self.y = y

        redraw_is_required = False
        if self.right:
            self.view.yaw += dx * 0.01
            self.view.pitch += dy * 0.01
            self.view.update_matrix()
            redraw_is_required = True

        if self.middle:
            plane_height = math.tan(
                self.projection.fov_y * 0.5) * self.view.distance * 2
            self.view.x += dx / self.height * plane_height
            self.view.y -= dy / self.height * plane_height
            self.view.update_matrix()
            redraw_is_required = True

        return redraw_is_required

    def onWheel(self, d: int) -> bool:
        ''' 
        Mouse input. Returns whether redraw is required.
        '''
        if d > 0:
            self.view.distance *= 1.1
            self.view.update_matrix()
            return True
        elif d < 0:
            self.view.distance *= 0.9
            self.view.update_matrix()
            return True
        return False

    def fit(self, p0: Float3, p1: Float3):
        if math.isnan(p0.x) or math.isnan(p0.y) or math.isnan(p0.z) or math.isnan(p1.x) or math.isnan(p1.y) or math.isnan(p1.z):
            return
        if math.isinf(p0.x) or math.isinf(p0.y) or math.isinf(p0.z) or math.isinf(p1.x) or math.isinf(p1.y) or math.isinf(p1.z):
            return

        self.view.x = 0
        self.view.y = -(p1.y+p0.y)/2
        self.view.distance = (p1.y-p0.y) / \
            math.tan(self.projection.fov_y / 2)
        self.view.yaw = 0
        self.view.pitch = 0
        self.view.update_matrix()
        logger.info(self.view)

        if self.view.distance*2 > self.projection.z_far:
            self.projection.z_far = self.view.distance*2
            self.projection.update_matrix()

    def get_mouse_ray(self):
        origin = Float3(
            self.view.inverse._41,
            self.view.inverse._42,
            self.view.inverse._43)
        half_fov = self.projection.fov_y/2
        dir = Float3(
            (self.x/self.width * 2 - 1) *
            math.tan(half_fov) * (self.projection.aspect),
            -(self.y/self.height * 2 - 1) * math.tan(half_fov),
            -1)
        dir = self.view.inverse.apply(*dir, translate=False)

        return Ray(origin, dir.normalized())

    def get_state(self, light: Float4) -> FrameState:
        ray = self.get_mouse_ray()
        return FrameState(
            Float4(0, 0, self.width, self.height),
            self.x, self.y,
            self.left, self.right, self.middle,
            self.view.matrix, self.projection.matrix,
            ray, light
        )



from typing import List, Iterable
import ctypes
import logging
#
from OpenGL import GL
import pydear as ImGui
from pydear.utils.dockspace import DockView
from glglue.basecontroller import BaseController
from .renderview import RenderView
logger = logging.getLogger(__name__)


def create_docks() -> List[DockView]:
    views: List[DockView] = []

    views.append(DockView(
        'metrics', (ctypes.c_bool * 1)(True), ImGui.ShowMetricsWindow))

    def show_hello(p_open: ctypes.Array):
        # open new window context
        if ImGui.Begin("CustomGUI", p_open):
            # draw text label inside of current window
            if ImGui.Button("Debug"):
                logger.debug("debug message")
        # close current window context
        ImGui.End()
    views.append(DockView(
        'hello', (ctypes.c_bool * 1)(True), show_hello))

    #
    # 3D View
    #
    render_view = RenderView()
    views.append(DockView(
        '3d', (ctypes.c_bool * 1)(True), render_view.draw))

    is_point = (ctypes.c_bool * 1)(False)

    def show_env(p_open: ctypes.Array):
        if ImGui.Begin("env", p_open):
            ImGui.Checkbox("point or direction", is_point)
            if is_point[0]:
                render_view.light.w = 1
            else:
                render_view.light.w = 0
            ImGui.SliderFloat3(
                'light position', render_view.light, -10, 10)  # type: ignore
        ImGui.End()
    views.append(DockView(
        'env', (ctypes.c_bool * 1)(True), show_env))

    # logger
    from pydear.utils.loghandler import ImGuiLogHandler
    log_handle = ImGuiLogHandler()
    log_handle.register_root()
    views.append(DockView('log', (ctypes.c_bool * 1)
                          (True), log_handle.draw))

    return views


class PydearController(BaseController):
    """
    [CLASSES] Controllerクラスは、glglueの規約に沿って以下のコールバックを実装する
    """

    def __init__(self, scale: float = 1):
        #
        # imgui
        #
        ImGui.CreateContext()
        self.io = ImGui.GetIO()
        self.io.ConfigFlags |= ImGui.ImGuiConfigFlags_.DockingEnable

        self.imgui_font()

        self.io.DisplayFramebufferScale = ImGui.ImVec2(scale, scale)
        from pydear.backends.opengl import Renderer
        self.impl_gl = Renderer()
        self.viewport = (1, 1)

        # imgui view
        self.imgui_docks = [dock for dock in self.imgui_create_docks()]

    def imgui_create_docks(self) -> List[DockView]:
        return create_docks()

    def imgui_font(self):
        # create texture before: ImGui.NewFrame()
        self.io.Fonts.GetTexDataAsRGBA32(
            (ctypes.c_void_p * 1)(),
            (ctypes.c_int * 1)(), (ctypes.c_int * 1)())

    def onResize(self, w, h):
        if self.viewport == (w, h):
            return False
        self.viewport = (w, h)
        return True

    def onLeftDown(self, x, y):
        self.io.MouseDown[0] = 1
        return False

    def onLeftUp(self, x, y):
        self.io.MouseDown[0] = 0
        return False

    def onMiddleDown(self, x, y):
        self.io.MouseDown[2] = 1
        return False

    def onMiddleUp(self, x, y):
        self.io.MouseDown[2] = 0
        return False

    def onRightDown(self, x, y):
        self.io.MouseDown[1] = 1
        return False

    def onRightUp(self, x, y):
        self.io.MouseDown[1] = 0
        return False

    def onMotion(self, x, y):
        self.io.MousePos = ImGui.ImVec2(x, y)
        return False

    def onWheel(self, d):
        self.io.MouseWheel = d
        return False

    def onKeyDown(self, keycode):
        logger.debug('onKeyDown: %d', keycode)

    def onUpdate(self, d):
        #logger.debug('onUpdate: delta %d ms', d)
        self.io.DeltaTime = d * 0.001
        return True

    def imgui_draw(self):
        from pydear.utils.dockspace import dockspace
        dockspace(self.imgui_docks)

    def draw(self):
        # state = self.camera.get_state()

        #
        # new frame
        #
        self.io.DisplaySize = ImGui.ImVec2(*self.viewport)
        ImGui.NewFrame()

        #
        # imgui
        #
        self.imgui_draw()
        ImGui.EndFrame()
        ImGui.Render()

        #
        # clear frame buffer
        #
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)
        GL.glViewport(0, 0, *self.viewport)
        GL.glClearColor(0.0, 0.0, 1.0, 0.0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT |
                   GL.GL_DEPTH_BUFFER_BIT)  # type: ignore

        #
        # render
        #
        # pass all drawing comands to the rendering pipeline
        # and close frame context
        self.impl_gl.render(ImGui.GetDrawData())

        GL.glFlush()



from typing import List, Optional, Union
from .material import Material
from .vertices import Planar, Interleaved, VectorView
from glglue import ctypesmath


class Submesh:
    def __init__(self, material: Material, macro: List[str], topology, offset: int, draw_count: int) -> None:
        self.material = material
        self.macro = macro

        self.topology = topology
        self.offset = offset
        self.draw_count = draw_count


class Mesh:
    def __init__(self, name: str, vertices: Union[Planar, Interleaved], indices: Optional[VectorView] = None) -> None:
        self.name = name
        self.indices = indices
        self.vertices = vertices
        self.submeshes: List[Submesh] = []
        # without skinning.
        self.aabb = ctypesmath.AABB.new_empty()

    def update(self, delta):
        pass

    def add_submesh(self, material: Material, macro: List[str], topology):
        draw_count = self.vertices.count()
        if self.indices:
            draw_count = self.indices.get_count()
        self.submeshes.append(
            Submesh(material, macro, topology, 0, draw_count))

    def expand_aabb(self, aabb):
        return aabb.expand(self.aabb)



import logging
import ctypes
from OpenGL import GL
from glglue import ctypesmath
from typing import Any, NamedTuple, List, Tuple
from .texture import Texture

logger = logging.getLogger(__name__)


def compile_shader(src: str, shader_type):
    shader = GL.glCreateShader(shader_type)
    GL.glShaderSource(shader, src)
    GL.glCompileShader(shader)
    result = GL.glGetShaderiv(shader, GL.GL_COMPILE_STATUS)
    if result != GL.GL_TRUE:
        info = GL.glGetShaderInfoLog(shader)
        GL.glDeleteShader(shader)
        logging.error(info)
        return
    return shader


class UniformMat4:
    def __init__(self, program, name: str) -> None:
        self.program = program
        self.name = name
        self.location = -1

    def set(self, value, transpose=True) -> None:
        if self.location < 0:
            self.location = GL.glGetUniformLocation(self.program, self.name)
        match value:
            case ctypesmath.Mat4():
                GL.glUniformMatrix4fv(
                    self.location, 1, GL.GL_TRUE if transpose else GL.GL_FALSE, ctypes.byref(value))
            case ctypesmath.Float4():
                GL.glUniform4fv(self.location, 1, ctypes.byref(value))
            case _:
                raise RuntimeError()


class ShaderSource(NamedTuple):
    vs: str
    vs_macro: Tuple[str, ...]
    fs: str
    fs_macro: Tuple[str, ...]

    def get_vs(self) -> str:
        return ''.join(f'{x}\n' for x in self.vs_macro) + self.vs

    def get_fs(self) -> str:
        return ''.join(f'{x}\n' for x in self.fs_macro) + self.fs


class Shader:
    def __init__(self) -> None:
        self.program = GL.glCreateProgram()
        self.ready = False
        self.uniforms = {}

    def __del__(self) -> None:
        GL.glDeleteProgram(self.program)

    def compile(self, shader_source: ShaderSource):
        vs = compile_shader(shader_source.get_vs(), GL.GL_VERTEX_SHADER)
        if not vs:
            logger.warn('compile_shader: GL_VERTEX_SHADER')
            return

        fs = compile_shader(shader_source.get_fs(), GL.GL_FRAGMENT_SHADER)
        if not fs:
            logging.warn('compile_shader: GL_FRAGMENT_SHADER')
            return

        GL.glAttachShader(self.program, vs)
        GL.glAttachShader(self.program, fs)
        GL.glLinkProgram(self.program)
        error = GL.glGetProgramiv(self.program, GL.GL_LINK_STATUS)
        GL.glDeleteShader(vs)
        GL.glDeleteShader(fs)
        if error != GL.GL_TRUE:
            info = GL.glGetShaderInfoLog(self.program)
            logger.error(info)

        self.ready = True

    def use(self):
        if not self.ready:
            return
        GL.glUseProgram(self.program)

    def unuse(self):
        GL.glUseProgram(0)

    def set_uniform(self, name: str, value: Any, transpose=True):
        if not self.ready:
            return
        u = self.uniforms.get(name)
        if not u:
            u = UniformMat4(self.program, name)
            self.uniforms[name] = u
        u.set(value, transpose)

    def set_texture(self, name: str, slot: int, texture: Texture):
        if not self.ready:
            return
        location = GL.glGetUniformLocation(self.program, name)
        texture.activate(location, slot)


def create_from(shader_source: ShaderSource) -> Shader:
    shader = Shader()
    shader.compile(shader_source)
    return shader



from typing import NamedTuple, Optional
import pathlib


class WindowConfig(NamedTuple):
    x: int
    y: int
    width: int
    height: int
    is_maximized: bool

    @staticmethod
    def load_json_from_path(file: pathlib.Path) -> Optional['WindowConfig']:
        if file.exists():
            import json
            return WindowConfig(**json.loads(file.read_bytes()))

    def save_json_to_path(self, file: pathlib.Path):
        with file.open('w') as w:
            import json
            json.dump(self._asdict(), w)



from typing import List, Optional, Union
from ..ctypesmath import *
from .mesh import Mesh
from .skin import Skin


class Node:
    def __init__(self, name: str, transform: Union[TRS, Mat4]):
        self. name = name
        self.local_transform = transform
        self.world_matrix = Mat4.new_identity()
        self.children: List['Node'] = []
        self.meshes: List[Mesh] = []
        self.skin: Optional[Skin] = None

    def update(self, delta):
        pass

    def get_local_matrix(self) -> Mat4:
        match self.local_transform:
            case Mat4() as m:
                return m
            case TRS() as trs:
                return trs.to_matrix()
            case _:
                raise RuntimeError()

    def expand_aabb(self, aabb: AABB, parent: Optional[Mat4] = None) -> AABB:
        if not parent:
            parent = Mat4.new_identity()
        m = self.get_local_matrix() * parent

        if self.meshes:
            local_aabb = AABB.new_empty()
            for mesh in self.meshes:
                local_aabb = local_aabb.expand(mesh.aabb)
            world_aabb = local_aabb.transform(m)
            aabb = aabb.expand(world_aabb)

        for child in self.children:
            aabb = child.expand_aabb(aabb, m)

        return aabb

    def calc_world(self, parent: Optional[Mat4] = None):
        if parent:
            self.world_matrix = self.get_local_matrix() * parent
        else:
            self.world_matrix = self.get_local_matrix()
        for child in self.children:
            child.calc_world(self.world_matrix)



import unittest
from glglue.ctypesmath import Float3


class TestCtypeMath(unittest.TestCase):

    def test_float3(self):
        self.assertEqual(Float3(2, 3, 4), Float3(1, 2, 3) + Float3(1, 1, 1))


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
from OpenGL.GL import *


class Coord(object):
    def __init__(self, size):
        self.size=size

    def draw(self):
        glBegin(GL_LINES)
        glColor(1, 0, 0); glVertex(0, 0, 0); glVertex(self.size, 0, 0)
        glColor(0.5, 0, 0); glVertex(0, 0, 0); glVertex(-self.size, 0, 0)
        glColor(0, 1, 0); glVertex(0, 0, 0); glVertex(0, self.size, 0)
        glColor(0, 0.5, 0); glVertex(0, 0, 0); glVertex(0, -self.size, 0)
        glColor(0, 0, 1); glVertex(0, 0, 0); glVertex(0, 0, self.size)
        glColor(0, 0, 0.5); glVertex(0, 0, 0); glVertex(0, 0, -self.size)
        glEnd()




# coding: utf-8

from abc import ABCMeta, abstractmethod


class BaseController(metaclass=ABCMeta):
    def __init__(self):
        pass

    @abstractmethod
    def onUpdate(*args) -> bool:
        """
        Update scene. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onLeftDown(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onLeftUp(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onMiddleDown(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onMiddleUp(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onRightDown(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onRightUp(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onMotion(self, x: int, y: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onResize(self, w: int, h: int) -> bool:
        """
        Gui event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onWheel(self, d: int) -> bool:
        """
        Mouse event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def onKeyDown(self, *args: str) -> bool:
        """
        Keyboard event. Returns whether redraw is required.
        """
        return False

    @abstractmethod
    def draw(self) -> None:
        """
        draw scene
        """
        pass
